#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include "anadisk.h"
#include <libdmk.h>
#include <dmk.h>

#define DEBUG_SCAN	1
#define DEBUG_FORMAT	1
#define DEBUG_WRITE	1
#define DEBUG_READ	1

FILE		*InFile;	// Anadisk dump file, including sector headers.
dmk_handle	OutFile;	// dmk file.

#define RPM		360
#define DataRate	500

int		MaxTrack;	// Maximum track number on disk
int		MaxHead;	// Maximum head on disk
int		MaxSector;	// Maximum sector no on disk
int		MinSector;	// Minimum sector no on disk
sector_mode_t	TrackDensity;	// Desnity to format tracks
sector_mode_t	Track0Density;	// Density to format track 0
int		Quiet;		// If true supress output (except errors)
int		t0s;		// Track 0 sectors

//
// Initialise globals.
//

void Init()
{
	MaxTrack=0;
	MaxHead=0;
	MaxSector=0;
	MinSector=255;
	TrackDensity=DMK_MFM;
	Track0Density=DMK_MFM;
	Quiet=0;
	t0s=-1;
}


//
// Print usage message.
//
void Usage()
{
	printf("ana2dmk, Copyright (C) 2005, P.Harvey-Smith.\n\n");
	printf("ana2dmk <anadisk_image> <dmk_iamge> [options]\n");
	printf("Options can be :-\n");
	printf("     -s       : single density disk\n");
	printf("     -t0s <n> : track 0 single density with n sectors, rest double\n");
	printf("     -q       : Quiet, supress output messages\n");
	printf("     -h       : Display this help\n");
	exit(0);
}

//
// Scan a sector read from anadisk file, and update MaxTrack,
// MaxHead, MaxSector and MinSector if needed. 
//
void ScanSector(ANADISK_SECTOR *sec)
{
	if (sec->LogicalCyl>MaxTrack)
		MaxTrack=sec->LogicalCyl;
		
	if (sec->LogicalHead>MaxHead)
		MaxHead=sec->LogicalHead;
		
	if (sec->LogicalSector>MaxSector)
		MaxSector=sec->LogicalSector;

	if (sec->LogicalSector<MinSector)
		MinSector=sec->LogicalSector;

#ifdef DEBUG_SCAN
	printf("Actual Cyl=%d, ",sec->ActualCyl);
	printf("Actual Head=%d, ",sec->ActualHead);
	printf("Log Cyl=%d, ",sec->LogicalCyl);
	printf("Log Head=%d, ",sec->LogicalHead);
	printf("Log Sector=%d, ",sec->LogicalSector);
	printf("Length=%d, ",sec->Length);
	printf("DataCount=%d\n",sec->DataCount);
#endif

}

//
// Return the density for the specified track.
//

sector_mode_t GetTrackDensity(int Track, int Head)
{
	if((Track==0) && (Head==0))
		return Track0Density;
	else
		return TrackDensity;
}

//
// Return sectors/track
//

int GetSectorsPerTrack(int Track, int Head)
{
	if((Track==0) && (Head==0))
		return t0s;
	else
		return MaxSector;
}

//
// Write an anadisk sector to DMK image
//

void WriteToDmk(ANADISK_SECTOR *sec, char *Buffer)
{
	sector_info_t	SectorInfo;
	
	SectorInfo.cylinder	= sec->LogicalCyl;
	SectorInfo.head		= sec->LogicalHead;
	SectorInfo.sector	= sec->LogicalSector;
	SectorInfo.size_code	= 1;
	SectorInfo.mode		= GetTrackDensity(sec->LogicalCyl,sec->LogicalHead);
	SectorInfo.write_data	= 0;
	SectorInfo.data_value	= 0xff;

#ifdef DEBUG_WRITE	
	printf("Write Sector C=%d H=%d R=%d\n",SectorInfo.cylinder,SectorInfo.head,SectorInfo.sector);
#endif

	dmk_seek(OutFile,SectorInfo.cylinder,SectorInfo.head);
	dmk_write_sector(OutFile,&SectorInfo,Buffer);
}

//
// Scan anadisk image, reads sequentially through all sector records in anadisk image, 
// and either calls ScanSector to update geometry values, or transfers the sector to
// the dmk image.
//

void ScanAnaImage(int transfer)
{
	ANADISK_SECTOR	SecHead;
	char		*Buffer;
	size_t		Count;
	
	fseek(InFile,0,SEEK_SET);
	
	Count=1;
	while (Count>0)
	{
		Count=fread(&SecHead,1,sizeof(SecHead),InFile);
		
#ifdef DEBUG_READ
		printf("Read %d bytes, at pos %d\n",Count,ftell(InFile));
#endif
		if (Count>0)
		{
			Buffer=(char *)malloc(SecHead.DataCount);	// Get a buffer for sector data
			fread(Buffer,SecHead.DataCount,1,InFile);	// Read from in, write to out
		
			if(transfer)
				WriteToDmk(&SecHead,Buffer);		// Write to DMK image
			else
				ScanSector(&SecHead);			// Update geometry values
			
			free(Buffer);					// Free buffer
		}
	}
}


//
// Format the DMK image, ittirates through tracks and heads formatting the track.
//

void FormatDMK(char *FileName)
{
	int		Track;
	int		Head;
	int		SectorNo;
	int		LastSector;
	sector_info_t 	sector_info [32];
	sector_mode_t	Mode;
	
	for(Track=0;Track<=MaxTrack;Track++)
	{
		for(Head=0;Head<=MaxHead;Head++)
		{
			if(!dmk_seek(OutFile,Track,Head))
			{
				printf("Error seeking to track:%d head:%d\n",Track,Head);
				exit(2);
			}
			
			Mode=GetTrackDensity(Track,Head);
			LastSector=GetSectorsPerTrack(Track,Head);
			for (SectorNo=MinSector; SectorNo<=LastSector;SectorNo++)
			{
				sector_info [SectorNo].cylinder   = Track;
				sector_info [SectorNo].head       = Head;
				sector_info [SectorNo].sector     = SectorNo;
				sector_info [SectorNo].size_code  = 1;
				sector_info [SectorNo].mode       = Mode;
				sector_info [SectorNo].write_data = 1;
				sector_info [SectorNo].data_value = 0xe5;  /* not used */
#ifdef DEBUG_FORMAT				
				printf("Format C=%d H=%d R=%d ",Track,Head,SectorNo);
				printf("Mode=%d Max-Min=%d\n",Mode, (LastSector-MinSector)+1);
#endif
			}	
			
			if (! dmk_format_track (OutFile, Mode, (LastSector-MinSector)+1, sector_info))
			{
				printf ("error formatting cylinder %d, head %d\n", Track, Head);
				exit (2);
			}
		}
	}
}

int main(int argc, char *argv[])
{
	int	Idx;
	
	Init();				// Init globals

	for(Idx=0;Idx<argc;Idx++)	// Decode command line params
	{
		if(!strcmp(argv[Idx],"-s"))
			TrackDensity=DMK_FM;
			
		if(!strcmp(argv[Idx],"-t0s"))
		{
			Track0Density=DMK_FM;			
			if(argc>=Idx+1)
				sscanf(argv[Idx+1],"%d",&t0s);
		}
		
		if(!strcmp(argv[Idx],"-q"))
			Quiet=1;
			
		if(!strcmp(argv[Idx],"-h"))
			Usage();						
	}

	if(argc<3)
		Usage();		// Display usage message
	else
	{		
		if((InFile=fopen(argv[1],"rb"))!=NULL)
		{
			if(!Quiet)
				printf("Scanning image : %s\n",argv[1]);
			
			ScanAnaImage(0);
			if(!Quiet)
				printf("MaxTrack=%d MaxHead=%d MaxSector=%d Density=%d Track 0 density=%d\n",MaxTrack,MaxHead,MaxSector,TrackDensity,Track0Density);

			if(t0s==-1)
				t0s=MaxSector;
		
			if ((OutFile=dmk_create_image(argv[2],MaxHead,MaxTrack+1,TrackDensity,RPM,DataRate))!=0)
			{
				if(!Quiet)
					printf("Formatting image : %s\n",argv[2]);
				
				FormatDMK(argv[2]);
				if(!Quiet)
					printf("Transfering.....\n");
				
				ScanAnaImage(1);
				dmk_close_image(OutFile);
			}
			else
				printf("Error opening output file : %s\n",argv[2]);
			
			fclose(InFile);
		}
		else
			printf("Error opening input file %s\n",argv[1]);
	}

	if(!Quiet)
		printf("Done !\n");
		
	return 0;
}
